package tracer;

import java.util.Iterator;

/**
 * Represents a ray: a ray can `trace' itself, calculating a color.
 */
public class Ray {

	public Vec3 origin;
	public Vec3 direction;
	
	public Ray( Vec3 o, Vec3 d ) {
		origin = new Vec3( o );
		direction = new Vec3( d );	
	}
	
	/**
	 * produceert de waarde van kleine omega in de perlin noise bereking
	 * @param t  de waarde waarvoor de kleine omega moet worden berekend
	 * @return  de waarde van kleine omega
	 */
	public float omega(float t)
	{
		float tabs = Math.abs(t); 
		if(tabs < 1 )
		{
			return (2*tabs*tabs*tabs) - (3*tabs*tabs) + 1;
		}
		else
			return 0f;
	}
	
	public int phi(int t, int n, Material mat)
	{
		/*
		 * hash functie gebruikt in perlin noise.
		 * Maakt dus gebruik van de p array, gemaakt in material.
		 */
		t = Math.abs(t);
		return mat.p[t % n];	
		
	}
	
	/**
	 * Calculates the local light term, given an IntersectionInfo and a Light.
	 * This method does <b>not</b> do an occlusion test; do this yourself (e.g.
	 * using a shadow feeler).
	 */
	public Vec3 localLight( IntersectionInfo info, Light light, Vec3 color ) 
	{
		// diffuse component
		Vec3 diffuse = color.times(light.color);
		Vec3 lightVec = light.location.minus(info.location);
		lightVec.normalize();
		diffuse = diffuse.times( Math.max(lightVec.dot(info.normal), 0) );
		diffuse = diffuse.times( info.object.material.diffuse * light.intensity );
		
		// specular (a.k.a. glossy) reflection
		Vec3 e = this.origin.minus( info.location );
		e.normalize();
		Vec3 l = light.location.minus( info.location );
		l.normalize();
		
		Vec3 h = e.add(l).times( 1 / (e.add(l).length()));
		float hdotnpowp = (float)Math.pow(h.dot(info.normal), info.object.material.specularPower);
		Vec3 specular = light.color.times( hdotnpowp ).times( info.object.material.specular);
		
		return diffuse.add( specular );
	}
	
	/**
	 * Does the actual `raytracing'. Returns the color this ray `hits.'
	 * @param currentObject This object is ignored in the intersection tests; in
	 *    effect, this object is `invisible' to the ray. This is useful for
	 *    reflection rays and shadow feelers: it avoids precision-errors by just
	 *    ignoring the object you've just bounced off of. If all objects are
	 *    convex (which they are in this tracer) this is actually not a hack but
	 *    completely correct.
	 * @param maxReflectionsLeft Maximum recursion depth for reflection
	 *    calculations .
	 */
	public Vec3 trace( Traceable currentObject, int maxReflectionsLeft ) {
		// test all Traceable object in the scene for collision.
		// store the nearest one.
		IntersectionInfo nearestHit = null;
		Iterator i = Tracer.scene.iterator();
		while( i.hasNext() ) {
			
			Traceable t = (Traceable)i.next();
			if( t != currentObject ) {
				IntersectionInfo info = t.intersect( this );
				if(  info.hit  &&  (nearestHit==null || info.distance<nearestHit.distance)  ) {
					nearestHit = info;
				}
			}
			
		}
		if( nearestHit != null ) {
			// actually hit something
			Material material = nearestHit.object.material;
			
			Vec3 color1 = textureMap(nearestHit);
			
			Vec3 color = color1.times( material.ambient ); // ambient
			
			// local contribution of light
			Iterator lightIter = Tracer.lights.iterator();
			while( lightIter.hasNext() ) {
				Light light = (Light)lightIter.next();
				Vec3 shadowFeelerDirection = light.location.minus( nearestHit.location );
				Ray shadowFeeler = new Ray( nearestHit.location, shadowFeelerDirection );
				if( !shadowFeeler.hit( nearestHit.object ) ) {
					color = color.add(  localLight( nearestHit, light, color1 )  );
				}
			}
			
			// global illumination: add recursively computed reflection below this line
			if(maxReflectionsLeft >= 0)
			{
				
				Vec3 r = this.origin.minus(nearestHit.location);
				r.normalize();
				Vec3 d = (r.times(-1)).add( nearestHit.normal.times ( nearestHit.normal.dot(r) * 2));
				Ray ray = new Ray(nearestHit.location, d);
				color = color.add(( ray.trace(nearestHit.object, maxReflectionsLeft-1)).times(nearestHit.object.material.reflectance));
							
			}
			return color;
						
		} else {
			// hit nothing; return background color
			return new Vec3( 0.0f, 0.0f, 0.0f );
		}
		
	}
	
	public Vec3 textureMap(IntersectionInfo info)
	{
		/*
		 * Texturemap wordt aangeroepen om de kleur van een object op een bepaalde
		 * intersectionpunt op te halen.
		 */
		
		Vec3 color = info.object.material.color;
		
		// perlin noise
		if(! info.object.material.color.equals(info.object.material.perlin))
		{
			//we moeten de perlin noise berekenen
			int floorx = (int) Math.floor(info.location.x);
			int floory = (int) Math.floor(info.location.y);
			int floorz = (int) Math.floor(info.location.z);
			float noise = 0;
			
			for(int i=floorx; i<= floorx+1; i++)
				for(int j=floory; j<= floory+1; j++)
					for(int k=floorz; k<=floorz+1; k++)
					{
						float u = info.location.x - i;
						float v = info.location.y - j;
						float w = info.location.z - k;
						int n = info.object.material.perlinArray.length;
						Vec3 randVec = info.object.material.perlinArray[phi(i + phi(j + phi(k,n,info.object.material), n,info.object.material), n,info.object.material)];
						noise += omega(u)*omega(v)*omega(w)*randVec.dot(new Vec3(u,v,w));
					}
			//manier van noise mappen naar 0-1 zoals in voorbeeldplaatje
			noise = Math.abs(noise);
			//andere manier van noise mappen naar 0-1
			//noise =(0.5f*noise) + 0.5f;
			
			color = (color.times(1-noise)).add(info.object.material.perlin.times(noise));
		}
		
		// texture map
		if(info.object.material.sampler != null)
			color = color.times( info.object.material.sampler.sample(info.u, info.v));
				
		return color;
	}
	
	/**
	 * Checks if the ray (origin + t*direction) hits the scene with 0 <= t <= 1.
	 * @param ignoreObject Similar to trace's currentObject parameter.
	 *    @see trace.
	 */
	public boolean hit( Traceable ignoreObject ) 
	{
		Iterator i = Tracer.scene.iterator();
		while( i.hasNext() ) 
		{
			Traceable t = (Traceable)i.next();
			if( t != ignoreObject ) 
			{
				if(t.hit( this ))
					return true;
			}
		}		
		return false;
	}
	
}